// This file is licensed under the Elastic License 2.0. Copyright 2021 StarRocks Limited.

#include "exec/vectorized/orc_scanner_adapter.h"

#include <gtest/gtest.h>

#include <ctime>
#include <filesystem>
#include <map>
#include <vector>

#include "common/object_pool.h"
#include "gen_cpp/Exprs_types.h"
#include "gutil/strings/substitute.h"
#include "runtime/descriptor_helper.h"
#include "runtime/descriptors.h"

namespace starrocks::vectorized {

class OrcScannerAdapterTest : public testing::Test {
public:
    OrcScannerAdapterTest();

protected:
    std::vector<SlotDescriptor*> _src_slot_descs;
    ObjectPool _pool;
};

struct SlotDesc {
    string name;
    TypeDescriptor type;
};

void create_slot_descriptors(ObjectPool* pool, std::vector<SlotDescriptor*>* res, SlotDesc* slot_descs, int size) {
    TDescriptorTableBuilder builder;
    TTupleDescriptorBuilder b3;
    DescriptorTbl* tbl;

    for (int i = 0; i < size; i++) {
        TSlotDescriptorBuilder b2;
        b2.column_name(slot_descs[i].name).type(slot_descs[i].type).id(i).nullable(true);
        b3.add_slot(b2.build());
    }
    b3.build(&builder);

    Status status = DescriptorTbl::create(pool, builder.desc_tbl(), &tbl);
    DCHECK(status.ok()) << status.get_error_msg();
    for (int i = 0; i < size; i++) {
        res->push_back(tbl->get_slot_descriptor(i));
    }
}

/**
 * orc-statistics /home/disk2/zy/tpch_10k.orc.zstd
 * 
File /home/disk2/zy/tpch_10k.orc.zstd has 7 columns
*** Column 0 ***
Column has 10000 values and has null value: no
*** Column 1 ***
Data type: Integer  Values: 10000 Has null: no Minimum: 1 Maximum: 7 Sum: 29963
*** Column 2 ***
Data type: Integer Values: 10000 Has null: no Minimum: 9 Maximum: 200000 Sum: 990162252
*** Column 3 ***
Data type: Integer Values: 10000 Has null: no Minimum: 1075520 Maximum: 1085632 Sum: 10806360843
*** Column 4 ***
Data type: String Values: 10000 Has null: no Minimum: 0.00 Maximum: 0.10 Total length: 40000
*** Column 5 ***
Data type: Integer Values: 10000 Has null: no Minimum: 1 Maximum: 50 Sum: 254871
*** Column 6 ***
Data type: Integer Values: 10000 Has null: no Minimum: 954 Maximum: 102549 Sum: 381929645

File /home/disk2/zy/tpch_10k.orc.zstd has 2 stripes
*** Stripe 0 ***

--- Column 0 ---
Column has 5120 values and has null value: no
--- Column 1 ---
Data type: Integer Values: 5120 Has null: no Minimum: 1 Maximum: 7 Sum: 15172
--- Column 2 ---
Data type: Integer Values: 5120 Has null: no Minimum: 9 Maximum: 199927 Sum: 507887212
--- Column 3 ---
Data type: Integer Values: 5120 Has null: no Minimum: 1075520 Maximum: 1080801 Sum: 5520239705
--- Column 4 ---
Data type: String Values: 5120 Has null: no Minimum: 0.00 Maximum: 0.10 Total length: 20480
--- Column 5 ---
Data type: Integer Values: 5120 Has null: no Minimum: 1 Maximum: 50 Sum: 129769
--- Column 6 ---
Data type: Integer Values: 5120 Has null: no Minimum: 1026 Maximum: 101380 Sum: 194471830

*** Stripe 1 ***

--- Column 0 ---
Column has 4880 values and has null value: no
--- Column 1 ---
Data type: Integer Values: 4880 Has null: no Minimum: 1 Maximum: 7 Sum: 14791
--- Column 2 ---
Data type: Integer Values: 4880 Has null: no Minimum: 19 Maximum: 200000 Sum: 482275040
--- Column 3 ---
Data type: Integer Values: 4880 Has null: no Minimum: 1080801 Maximum: 1085632 Sum: 5286121138
--- Column 4 ---
Data type: String Values: 4880 Has null: no Minimum: 0.00 Maximum: 0.10 Total length: 19520
--- Column 5 ---
Data type: Integer Values: 4880 Has null: no Minimum: 1 Maximum: 50 Sum: 125102
--- Column 6 ---
Data type: Integer Values: 4880 Has null: no Minimum: 954 Maximum: 102549 Sum: 187457815
*/

static const std::string input_orc_file = "./be/test/exec/test_data/orc_scanner/tpch_10k.orc.zstd";
static const uint64_t total_record_num = 10 * 1000;
static const uint64_t default_row_group_size = 1000;
/** 
 * orc-metadata /home/disk2/zy/tpch_10k.orc.zstd
 * 
{ "name": "/home/disk2/zy/tpch_10k.orc.zstd",
  "type": "struct<lo_custkey:tinyint,lo_orderdate:int,lo_orderkey:int,lo_orderpriority:string,lo_partkey:tinyint,lo_suppkey:int>",
  "attributes": {},
  "rows": 10000,
  "stripe count": 2,
  "format": "0.12", "writer version": "future - 9",
  "compression": "zstd", "compression block": 4096,
  "file length": 76791,
  "content": 76199, "stripe stats": 268, "footer": 299, "postscript": 24,
  "row index stride": 1000,
  "user metadata": {
  },
  "stripes": [
    { "stripe": 0, "rows": 5120,
      "offset": 3, "length": 39170,
      "index": 850, "data": 38187, "footer": 133
    },
    { "stripe": 1, "rows": 4880,
      "offset": 39173, "length": 37026,
      "index": 739, "data": 36156, "footer": 131
    }
  ]
}
*/

OrcScannerAdapterTest::OrcScannerAdapterTest() {
    SlotDesc slot_descs[] = {
            {"lo_custkey", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_TINYINT)},
            {"lo_orderdate", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT)},
            {"lo_orderkey", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT)},
            {"lo_orderpriority", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_VARCHAR)},
            {"lo_partkey", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_TINYINT)},
            {"lo_suppkey", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT)},
    };
    const int n = sizeof(slot_descs) / sizeof(slot_descs[0]);
    create_slot_descriptors(&_pool, &_src_slot_descs, slot_descs, n);
}

static uint64_t get_hit_rows(OrcScannerAdapter* adapter) {
    uint64_t records = 0;
    for (;;) {
        Status st = adapter->read_next();
        if (st.is_end_of_file()) {
            break;
        }
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr ckptr = adapter->create_chunk();
        DCHECK(ckptr != nullptr);
        st = adapter->fill_chunk(&ckptr);
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr result = adapter->cast_chunk(&ckptr);
        DCHECK(result != nullptr);
        records += result->num_rows();
        DCHECK(result->num_columns() == adapter->num_columns());
    }
    return records;
}

TEST_F(OrcScannerAdapterTest, Normal) {
    OrcScannerAdapter adapter(_src_slot_descs);
    auto input_stream = orc::readLocalFile(input_orc_file);
    adapter.init(std::move(input_stream));
    uint64_t records = get_hit_rows(&adapter);
    EXPECT_EQ(records, total_record_num);
}

class SkipStripeRowFilter : public orc::RowReaderFilter {
public:
    bool filterOnOpeningStripe(uint64_t stripIndex, const orc::proto::StripeInformation* stripeInformation) override {
        // we are going to read strip 1
        // and there are 4880 rows
        if (stripIndex == 1) {
            return false;
        }
        return true;
    }
    uint64_t expected_rows() const { return 4880; }
};

TEST_F(OrcScannerAdapterTest, SkipStripe) {
    OrcScannerAdapter adapter(_src_slot_descs);
    auto filter = std::make_shared<SkipStripeRowFilter>();
    adapter.set_row_reader_filter(filter);

    auto input_stream = orc::readLocalFile(input_orc_file);
    adapter.init(std::move(input_stream));

    uint64_t records = get_hit_rows(&adapter);
    EXPECT_EQ(records, filter->expected_rows());
}

static TTypeDesc create_primitive_type_desc(TPrimitiveType::type type) {
    TTypeDesc result;
    TTypeNode node;
    node.__set_type(TTypeNodeType::SCALAR);
    TScalarType scalar_type;
    scalar_type.__set_type(type);
    node.__set_scalar_type(scalar_type);
    result.types.push_back(node);
    return result;
}

static TExprNode create_int_literal_node(TPrimitiveType::type value_type, int64_t value_literal) {
    TExprNode lit_node;
    lit_node.__set_node_type(TExprNodeType::INT_LITERAL);
    lit_node.__set_num_children(0);
    lit_node.__set_type(create_primitive_type_desc(value_type));
    TIntLiteral lit_value;
    lit_value.__set_value(value_literal);
    lit_node.__set_int_literal(lit_value);
    lit_node.__set_use_vectorized(true);
    return lit_node;
}

[[maybe_unused]] static TExprNode create_string_literal_node(TPrimitiveType::type value_type,
                                                             const std::string& value_literal) {
    TExprNode lit_node;
    lit_node.__set_node_type(TExprNodeType::STRING_LITERAL);
    lit_node.__set_num_children(0);
    lit_node.__set_type(create_primitive_type_desc(value_type));
    TStringLiteral lit_value;
    lit_value.__set_value(value_literal);
    lit_node.__set_string_literal(lit_value);
    lit_node.__set_use_vectorized(true);
    return lit_node;
}

template <typename ValueType>
static void push_binary_pred_texpr_node(std::vector<TExprNode>& nodes, TExprOpcode::type opcode,
                                        SlotDescriptor* slot_desc, ValueType value_type, TExprNode lit_node) {
    TExprNode eq_node;
    eq_node.__set_node_type(TExprNodeType::type::BINARY_PRED);
    eq_node.__set_child_type(value_type);
    eq_node.__set_type(create_primitive_type_desc(TPrimitiveType::BOOLEAN));
    eq_node.__set_opcode(opcode);
    eq_node.__set_num_children(2);
    eq_node.__set_use_vectorized(true);

    TExprNode slot_node;
    slot_node.__set_node_type(TExprNodeType::SLOT_REF);
    slot_node.__set_num_children(0);
    slot_node.__set_type(create_primitive_type_desc(value_type));
    TSlotRef slot_ref;
    slot_ref.__set_slot_id(slot_desc->id());
    slot_ref.__set_tuple_id(slot_desc->parent());
    slot_node.__set_slot_ref(slot_ref);
    slot_node.__set_use_vectorized(true);

    nodes.emplace_back(eq_node);
    nodes.emplace_back(slot_node);
    nodes.emplace_back(lit_node);
}

template <typename ValueType>
static void push_inpred_texpr_node(std::vector<TExprNode>& nodes, TExprOpcode::type opcode, SlotDescriptor* slot_desc,
                                   ValueType value_type, std::vector<TExprNode>& lit_nodes) {
    TExprNode eq_node;
    // vectorized does not support is null pred
    // eq_node.__set_node_type(TExprNodeType::type::IS_NULL_PRED);
    eq_node.__set_node_type(TExprNodeType::type::IN_PRED);
    eq_node.__set_opcode(opcode);
    eq_node.__set_child_type(value_type);
    eq_node.__set_type(create_primitive_type_desc(TPrimitiveType::BOOLEAN));
    eq_node.__set_num_children(1 + lit_nodes.size());
    eq_node.__set_use_vectorized(true);

    TExprNode slot_node;
    slot_node.__set_node_type(TExprNodeType::SLOT_REF);
    slot_node.__set_num_children(0);
    slot_node.__set_type(create_primitive_type_desc(value_type));
    TSlotRef slot_ref;
    slot_ref.__set_slot_id(slot_desc->id());
    slot_ref.__set_tuple_id(slot_desc->parent());
    slot_node.__set_slot_ref(slot_ref);
    slot_node.__set_use_vectorized(true);

    nodes.emplace_back(eq_node);
    nodes.emplace_back(slot_node);

    for (TExprNode& lit_node : lit_nodes) {
        nodes.emplace_back(lit_node);
    }
}

static ExprContext* create_expr_context(ObjectPool* pool, const std::vector<TExprNode>& nodes) {
    TExpr texpr;
    texpr.__set_nodes(nodes);
    ExprContext* ctx;
    Status st = Expr::create_expr_tree(pool, texpr, &ctx);
    DCHECK(st.ok()) << st.get_error_msg();
    return ctx;
}

TEST_F(OrcScannerAdapterTest, SkipFileByConjunctsEQ) {
    OrcScannerAdapter adapter(_src_slot_descs);

    // lo_custkey == 0, min/max is 1,7.
    std::vector<TExprNode> nodes;
    int slot_index = 0;
    TExprNode lit_node = create_int_literal_node(TPrimitiveType::TINYINT, 0);
    push_binary_pred_texpr_node(nodes, TExprOpcode::type::EQ, _src_slot_descs[slot_index], TPrimitiveType::TINYINT,
                                lit_node);
    ExprContext* ctx = create_expr_context(&_pool, nodes);
    std::vector<Expr*> conjuncts = {ctx->root()};
    adapter.set_conjuncts(conjuncts);

    auto input_stream = orc::readLocalFile(input_orc_file);
    adapter.init(std::move(input_stream));
    uint64_t records = get_hit_rows(&adapter);
    EXPECT_EQ(records, 0);
}

TEST_F(OrcScannerAdapterTest, SkipStripeByConjunctsEQ) {
    OrcScannerAdapter adapter(_src_slot_descs);

    // lo_orderdate == 200000
    // stripe0 min/max = 9/199927 [5120]
    // stripe1 min/max= 19/200000 [4880]
    std::vector<TExprNode> nodes;
    int slot_index = 1;
    TExprNode lit_node = create_int_literal_node(TPrimitiveType::INT, 200000);
    push_binary_pred_texpr_node(nodes, TExprOpcode::type::EQ, _src_slot_descs[slot_index], TPrimitiveType::INT,
                                lit_node);
    ExprContext* ctx = create_expr_context(&_pool, nodes);
    std::vector<Expr*> conjuncts = {ctx->root()};
    adapter.set_conjuncts(conjuncts);

    auto input_stream = orc::readLocalFile(input_orc_file);
    adapter.init(std::move(input_stream));
    uint64_t records = get_hit_rows(&adapter);
    // at most read one stripe.
    EXPECT_LE(records, 4880);
    // and actually just read one row group.
    // because there is only one item value is 200000
    // and that item belongs to one row group.
    EXPECT_EQ(records, default_row_group_size);
}

TEST_F(OrcScannerAdapterTest, SkipStripeByConjunctsInPred) {
    OrcScannerAdapter adapter(_src_slot_descs);

    // lo_orderdate min/max = 9/200000
    std::vector<TExprNode> nodes;
    int slot_index = 1;
    std::vector<int64_t> values = {0, 1, 2, 3, 4, 200001, 200002};
    std::vector<TExprNode> lit_nodes;
    for (int64_t v : values) {
        auto lit_node = create_int_literal_node(TPrimitiveType::INT, v);
        lit_nodes.emplace_back(lit_node);
    }
    push_inpred_texpr_node(nodes, TExprOpcode::type::FILTER_IN, _src_slot_descs[slot_index], TPrimitiveType::INT,
                           lit_nodes);
    ExprContext* ctx = create_expr_context(&_pool, nodes);
    std::vector<Expr*> conjuncts = {ctx->root()};
    adapter.set_conjuncts(conjuncts);

    auto input_stream = orc::readLocalFile(input_orc_file);
    adapter.init(std::move(input_stream));
    uint64_t records = get_hit_rows(&adapter);
    EXPECT_LE(records, 0);
}

class SkipRowGroupRowFilter : public orc::RowReaderFilter {
public:
    bool filterOnOpeningStripe(uint64_t stripeIndex, const orc::proto::StripeInformation* stripeInformation) override {
        currentStripeIndex = stripeIndex;
        return false;
    }

    bool filterOnPickRowGroup(size_t rowGroupIdx, const std::unordered_map<uint64_t, orc::proto::RowIndex>& rowIndexes,
                              const std::map<uint32_t, orc::BloomFilterIndex>& bloomFilters) override {
        // only reads first two row group2 of first stripe
        if (rowGroupIdx < 2 && currentStripeIndex == 1) {
            return false;
        }
        return true;
    }
    void onStartingPickRowGroups() override {
        std::cout << "on starting pick row groups of stripe = " << currentStripeIndex << std::endl;
    }
    void onEndingPickRowGroups() override {
        std::cout << "on ending pick row groups of stripe = " << currentStripeIndex << std::endl;
    }
    uint64_t expected_rows() const { return default_row_group_size * 2; }

private:
    uint64_t currentStripeIndex;
};

TEST_F(OrcScannerAdapterTest, SkipRowGroups) {
    OrcScannerAdapter adapter(_src_slot_descs);
    auto filter = std::make_shared<SkipRowGroupRowFilter>();
    adapter.set_row_reader_filter(filter);

    auto input_stream = orc::readLocalFile(input_orc_file);
    adapter.init(std::move(input_stream));

    uint64_t records = get_hit_rows(&adapter);
    EXPECT_EQ(records, filter->expected_rows());
}

template <int ORC_PRECISION, int ORC_SCALE, typename ValueType>
std::vector<DecimalV2Value> convert_orc_to_starrocks_decimalv2(ObjectPool* pool, const std::vector<ValueType>& values) {
    std::cout << "orc precision=" << ORC_PRECISION << " scale=" << ORC_SCALE << std::endl;
    if constexpr (std::is_same_v<ValueType, int64_t>) {
        static_assert(ORC_PRECISION <= 18);
    } else {
        static_assert(std::is_same_v<ValueType, int128_t>);
        static_assert(ORC_PRECISION > 18);
    }

    using OrcDecimalVectorBatch =
            std::conditional_t<ORC_PRECISION <= 18, orc::Decimal64VectorBatch, orc::Decimal128VectorBatch>;

    const char* filename = "orc_scanner_test_decimal.orc";
    std::filesystem::remove(filename);
    ORC_UNIQUE_PTR<orc::OutputStream> outStream = orc::writeLocalFile(filename);
    ORC_UNIQUE_PTR<orc::Type> schema(
            orc::Type::buildTypeFromString(strings::Substitute("struct<c0:decimal($0,$1)>", ORC_PRECISION, ORC_SCALE)));
    ORC_UNIQUE_PTR<orc::Writer> writer = createWriter(*schema, outStream.get(), orc::WriterOptions{});

    CHECK_LE(values.size(), 1024);
    ORC_UNIQUE_PTR<orc::ColumnVectorBatch> batch = writer->createRowBatch(1024);
    auto* root = dynamic_cast<orc::StructVectorBatch*>(batch.get());
    auto* c0 = dynamic_cast<OrcDecimalVectorBatch*>(root->fields[0]);

    if constexpr (std::is_same_v<ValueType, int64_t>) {
        for (int i = 0; i < values.size(); i++) {
            c0->values[i] = values[i];
        }
    } else {
        for (int i = 0; i < values.size(); i++) {
            c0->values[i] = orc::Int128(values[i] >> 64, values[i]);
        }
    }
    root->numElements = values.size();
    c0->numElements = values.size();
    writer->add(*batch);
    writer->close();
    outStream->close();

    ORC_UNIQUE_PTR<orc::Reader> reader = orc::createReader(orc::readLocalFile(filename), orc::ReaderOptions{});

    TDescriptorTableBuilder builder;
    TTupleDescriptorBuilder b3;

    TypeDescriptor type(TYPE_DECIMALV2);
    type.precision = 27;
    type.scale = 9;
    DescriptorTbl* tbl;
    std::vector<SlotDescriptor*> slots;
    TSlotDescriptorBuilder b2;
    b2.column_name("c0").type(type).id(0).nullable(true);
    b3.add_slot(b2.build());
    b3.build(&builder);

    Status status = DescriptorTbl::create(pool, builder.desc_tbl(), &tbl);
    DCHECK(status.ok()) << status.get_error_msg();
    slots.push_back(tbl->get_slot_descriptor(0));

    OrcScannerAdapter adapter(slots);
    adapter.init(std::move(reader));
    Status st = adapter.read_next();
    CHECK(st.ok()) << st.to_string();

    auto chunk = std::make_shared<Chunk>();
    chunk->append_column(ColumnHelper::create_column(type, true), 0);

    st = adapter.fill_chunk(&chunk);
    CHECK(st.ok()) << st.to_string();

    st = adapter.read_next();
    CHECK(!st.ok());
    std::filesystem::remove(filename);

    auto nullable = std::static_pointer_cast<NullableColumn>(chunk->get_column_by_index(0));
    CHECK(!nullable->has_null());
    auto decimal_col = std::static_pointer_cast<DecimalColumn>(nullable->data_column());
    return decimal_col->get_data();
}

TEST_F(OrcScannerAdapterTest, TestDecimal64) {
    const std::vector<int64_t> orc_values = {
            0,
            1,
            -1,
            123,
            123'000'000'000,
            -999'999'999,
            -999'999'999'999'999'999,
            999'999'999,
            999'999'999'999'999'999,
            1'000'000'000,
            999'999'999'000'000'000,
    };

    auto check_results = [](const std::vector<const char*>& exp, const std::vector<DecimalV2Value>& real) {
        ASSERT_EQ(exp.size(), real.size());
        for (int i = 0; i < exp.size(); i++) {
            EXPECT_EQ(exp[i], real[i].to_string());
        }
    };

    ObjectPool pool;
    {
        std::vector<const char*> exp = {
                "0",
                "0.000000001",
                "-0.000000001",
                "0.000000123",
                "123",
                "-0.999999999",
                "-999999999.999999999",
                "0.999999999",
                "999999999.999999999",
                "1",
                "999999999",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 9>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",           "0.00000001",           "-0.00000001", "0.00000123",          "1230",
                "-9.99999999", "-9999999999.99999999", "9.99999999",  "9999999999.99999999", "10",
                "9999999990",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 8>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0.0000001",
                "-0.0000001",
                "0.0000123",
                "12300",
                "-99.9999999",
                "-99999999999.9999999",
                "99.9999999",
                "99999999999.9999999",
                "100",
                "99999999900",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 7>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "1",
                "-1",
                "123",
                "123000000000",
                "-999999999",
                "-999999999999999999",
                "999999999",
                "999999999999999999",
                "1000000000",
                "999999999000000000",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 0>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0",
                "0",
                "0.000000012",
                "12.3",
                "-0.099999999",
                "-99999999.999999999",
                "0.099999999",
                "99999999.999999999",
                "0.1",
                "99999999.9",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 10>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0",
                "0",
                "0.000000001",
                "1.23",
                "-0.009999999",
                "-9999999.999999999",
                "0.009999999",
                "9999999.999999999",
                "0.01",
                "9999999.99",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 11>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",           "0",           "0",          "0",          "0.00000123", "-0.000000009", "-9.999999999",
                "0.000000009", "9.999999999", "0.00000001", "9.99999999",
        };
        auto real = convert_orc_to_starrocks_decimalv2<18, 17>(&pool, orc_values);
        check_results(exp, real);
    }
}

TEST_F(OrcScannerAdapterTest, TestDecimal128) {
    const std::vector<int128_t> orc_values = {
            (int128_t)0,
            (int128_t)1,
            (int128_t)-1,
            (int128_t)123,
            (int128_t)123'000'000'000,
            (int128_t)-999'999'999,
            // -999'999'999'999'999'999'999'999'999
            -((int128_t)999'999'999'999'999'999 * (int128_t)1'000'000'000 + (int128_t)999'999'999),
            (int128_t)999'999'999,
            // +999'999'999'999'999'999'999'999'999
            (int128_t)999'999'999'999'999'999 * (int128_t)1'000'000'000 + (int128_t)999'999'999,
            (int128_t)1'000'000'000,
            (int128_t)999'999'999'000'000'000,
    };

    auto check_results = [](const std::vector<const char*>& exp, const std::vector<DecimalV2Value>& real) {
        ASSERT_EQ(exp.size(), real.size());
        for (int i = 0; i < exp.size(); i++) {
            EXPECT_EQ(exp[i], real[i].to_string());
        }
    };

    ObjectPool pool;
    {
        std::vector<const char*> exp = {
                "0",
                "0.000000001",
                "-0.000000001",
                "0.000000123",
                "123",
                "-0.999999999",
                "-999999999999999999.999999999",
                "0.999999999",
                "999999999999999999.999999999",
                "1",
                "999999999",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 9>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0.00000001",
                "-0.00000001",
                "0.00000123",
                "1230",
                "-9.99999999",
                "-9999999999999999999.99999999",
                "9.99999999",
                "9999999999999999999.99999999",
                "10",
                "9999999990",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 8>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0.0000001",
                "-0.0000001",
                "0.0000123",
                "12300",
                "-99.9999999",
                "-99999999999999999999.9999999",
                "99.9999999",
                "99999999999999999999.9999999",
                "100",
                "99999999900",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 7>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "1",
                "-1",
                "123",
                "123000000000",
                "-999999999",
                "-999999999999999999999999999",
                "999999999",
                "999999999999999999999999999",
                "1000000000",
                "999999999000000000",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 0>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0",
                "0",
                "0.000000012",
                "12.3",
                "-0.099999999",
                "-99999999999999999.999999999",
                "0.099999999",
                "99999999999999999.999999999",
                "0.1",
                "99999999.9",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 10>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0",
                "0",
                "0",
                "0.000000001",
                "1.23",
                "-0.009999999",
                "-9999999999999999.999999999",
                "0.009999999",
                "9999999999999999.999999999",
                "0.01",
                "9999999.99",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 11>(&pool, orc_values);
        check_results(exp, real);
    }
    {
        std::vector<const char*> exp = {
                "0", "0", "0", "0", "0", "0", "-9.999999999", "0", "9.999999999", "0", "0.000000009",
        };
        auto real = convert_orc_to_starrocks_decimalv2<27, 26>(&pool, orc_values);
        check_results(exp, real);
    }
}

std::vector<TimestampValue> convert_orc_to_starrocks_timestamp(ObjectPool* pool, const std::string& reader_tz,
                                                               const std::string& write_tz,
                                                               const std::vector<int64_t>& values) {
    const char* filename = "orc_scanner_test_timestamp.orc";
    std::filesystem::remove(filename);
    ORC_UNIQUE_PTR<orc::OutputStream> outStream = orc::writeLocalFile(filename);
    ORC_UNIQUE_PTR<orc::Type> schema(orc::Type::buildTypeFromString("struct<c0:timestamp>"));

    orc::WriterOptions writer_options;
    writer_options.setTimezoneName(write_tz);
    ORC_UNIQUE_PTR<orc::Writer> writer = createWriter(*schema, outStream.get(), writer_options);

    CHECK_LE(values.size(), 1024);
    ORC_UNIQUE_PTR<orc::ColumnVectorBatch> batch = writer->createRowBatch(1024);
    auto* root = dynamic_cast<orc::StructVectorBatch*>(batch.get());
    auto* c0 = dynamic_cast<orc::TimestampVectorBatch*>(root->fields[0]);

    for (int i = 0; i < values.size(); i++) {
        c0->data[i] = values[i];
        c0->nanoseconds[i] = 0;
    }

    root->numElements = values.size();
    c0->numElements = values.size();
    writer->add(*batch);
    writer->close();
    outStream->close();

    ORC_UNIQUE_PTR<orc::Reader> reader = orc::createReader(orc::readLocalFile(filename), orc::ReaderOptions{});

    TDescriptorTableBuilder builder;
    TTupleDescriptorBuilder b3;

    TypeDescriptor type(TYPE_DATETIME);
    DescriptorTbl* tbl;
    std::vector<SlotDescriptor*> slots;
    TSlotDescriptorBuilder b2;
    b2.column_name("c0").type(type).id(0).nullable(true);
    b3.add_slot(b2.build());
    b3.build(&builder);

    Status status = DescriptorTbl::create(pool, builder.desc_tbl(), &tbl);
    DCHECK(status.ok()) << status.get_error_msg();
    slots.push_back(tbl->get_slot_descriptor(0));

    OrcScannerAdapter adapter(slots);
    adapter.set_timezone(reader_tz);
    adapter.init(std::move(reader));
    Status st = adapter.read_next();
    CHECK(st.ok()) << st.to_string();

    auto chunk = std::make_shared<Chunk>();
    chunk->append_column(ColumnHelper::create_column(type, true), 0);

    st = adapter.fill_chunk(&chunk);
    CHECK(st.ok()) << st.to_string();

    st = adapter.read_next();
    CHECK(!st.ok());
    std::filesystem::remove(filename);

    auto nullable = std::static_pointer_cast<NullableColumn>(chunk->get_column_by_index(0));
    CHECK(!nullable->has_null());
    auto ts_col = std::static_pointer_cast<TimestampColumn>(nullable->data_column());
    return ts_col->get_data();
}

TEST_F(OrcScannerAdapterTest, TestTimestamp) {
    // clang-format off
    const std::vector<int64_t> orc_values = {
        // 2021.5.25 1:18:40 GMT
        // 2021.5.25 9:18:40 Asia/Shanghai
        1621905520,
        // 1970.1.1 0:0:0 GMT
        // 1970.1.1 8:00:0 Asia/Shanghai
        0,
        // 1702.5.31 19:55:52 GMT
        // to Asia/Shanghai, it's supposed to be
        // 1702.6.01 03:55:52
        // but acutally it's
        // 1702-06-01 04:01:35
        // before unix epoch, time conversion is totoally a mess.
        -8444232248
    };
    const std::vector<std::string> exp_values = {
        "2021-05-25 09:18:40",
        "1970-01-01 08:00:00",
        "1702-06-01 04:01:35",
    };
    // clang-format on
    ObjectPool pool;
    auto res = convert_orc_to_starrocks_timestamp(&pool, "Asia/Shanghai", "UTC", orc_values);
    EXPECT_EQ(res.size(), orc_values.size());
    for (size_t i = 0; i < res.size(); i++) {
        std::string o = res[i].to_string();
        EXPECT_EQ(o, exp_values[i]);
    }
}

/**
 { "name": "StarRocks/be/test/exec/test_data/orc_scanner/orc_test_positional_column.orc",
  "type": "struct<a:int,b:string,_col12:int,c:double>",
  "attributes": {},
  "rows": 2,
  "stripe count": 1,
  "format": "0.12", "writer version": "ORC-135",
  "compression": "zlib", "compression block": 65536,
  "file length": 566,
  "content": 308, "stripe stats": 81, "footer": 149, "postscript": 24,
  "row index stride": 10000,
  "user metadata": {
  },
  "stripes": [
    { "stripe": 0, "rows": 2,
      "offset": 3, "length": 308,
      "index": 146, "data": 77, "footer": 85
    }
  ]
}

{"a": 10, "b": "hello", "_col12": 30, "c": 10.01}
{"a": 20, "b": "world", "_col12": 40, "c": 20.01}
 */

TEST_F(OrcScannerAdapterTest, TestReadPositionalColumn) {
    SlotDesc slot_descs[] = {
            {"a", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT)},
            {"b", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_VARCHAR)},
            {"c", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_DOUBLE)},
    };
    static const std::string input_orc_file = "./be/test/exec/test_data/orc_scanner/orc_test_positional_column.orc";
    std::vector<SlotDescriptor*> src_slot_descriptors;
    const int n = sizeof(slot_descs) / sizeof(slot_descs[0]);
    ObjectPool pool;
    create_slot_descriptors(&pool, &src_slot_descriptors, slot_descs, n);

    {
        OrcScannerAdapter adapter(src_slot_descriptors);
        auto input_stream = orc::readLocalFile(input_orc_file);
        Status st = adapter.init(std::move(input_stream));
        DCHECK(st.ok()) << st.get_error_msg();

        st = adapter.read_next();
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr ckptr = adapter.create_chunk();
        DCHECK(ckptr != nullptr);
        st = adapter.fill_chunk(&ckptr);
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr result = adapter.cast_chunk(&ckptr);
        DCHECK(result != nullptr);

        EXPECT_EQ(result->num_rows(), 2);
        EXPECT_EQ(result->num_columns(), 3);

        // slot_id = 0 is column a
        // we read 10, 20
        ColumnPtr col = result->get_column_by_slot_id(0);
        EXPECT_FALSE(col->is_null(0));
        EXPECT_EQ(col->get(0).get_int32(), 10);
        EXPECT_FALSE(col->is_null(1));
        EXPECT_EQ(col->get(1).get_int32(), 20);
    }

    {
        OrcScannerAdapter adapter(src_slot_descriptors);
        std::vector<std::string> hive_column_names = {"mm", "b", "a", "c"};
        adapter.set_hive_column_names(&hive_column_names);
        auto input_stream = orc::readLocalFile(input_orc_file);
        Status st = adapter.init(std::move(input_stream));
        DCHECK(st.ok()) << st.get_error_msg();

        st = adapter.read_next();
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr ckptr = adapter.create_chunk();
        DCHECK(ckptr != nullptr);
        st = adapter.fill_chunk(&ckptr);
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr result = adapter.cast_chunk(&ckptr);
        DCHECK(result != nullptr);

        EXPECT_EQ(result->num_rows(), 2);
        EXPECT_EQ(result->num_columns(), 3);

        // slot_id = 0 is column a
        // since we have hive column names
        // we can do positional read.
        // so content is 30, 40
        ColumnPtr col = result->get_column_by_slot_id(0);
        EXPECT_FALSE(col->is_null(0));
        EXPECT_EQ(col->get(0).get_int32(), 30);
        EXPECT_FALSE(col->is_null(1));
        EXPECT_EQ(col->get(1).get_int32(), 40);
    }
}

/**
 *
{ "name": "/home/disk2/zy/StarRocks/be/test/exec/test_data/orc_scanner/orc_test_array_basic.orc",
  "type": "struct<c0:int,c1:array<int>,c2:array<int>,c3:int,c4:double>",
  "attributes": {},
  "rows": 2,
  "stripe count": 1,
  "format": "0.12", "writer version": "ORC-135",
  "compression": "zlib", "compression block": 65536,
  "file length": 689,
  "content": 399, "stripe stats": 92, "footer": 170, "postscript": 24,
  "row index stride": 10000,
  "user metadata": {
  },
  "stripes": [
    { "stripe": 0, "rows": 2,
      "offset": 3, "length": 399,
      "index": 195, "data": 105, "footer": 99
    }
  ]
}

{"c0": 10, "c1": [10, 20, 30], "c2": [30, 40, 50], "c3": 60, "c4": 60.1}
{"c0": 20, "c1": [20, 30, 40], "c2": [40, 50, 60], "c3": 70, "c4": 70.2}
 */

TEST_F(OrcScannerAdapterTest, TestReadArrayBasic) {
    SlotDesc slot_descs[] = {
            {"c0", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT)},
            {"c1", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_ARRAY)},
            {"c2", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_ARRAY)},
            {"c4", TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_DOUBLE)},
    };

    slot_descs[1].type.children.push_back(TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT));
    slot_descs[2].type.children.push_back(TypeDescriptor::from_primtive_type(PrimitiveType::TYPE_INT));

    static const std::string input_orc_file = "./be/test/exec/test_data/orc_scanner/orc_test_array_basic.orc";
    std::vector<SlotDescriptor*> src_slot_descriptors;
    const int n = sizeof(slot_descs) / sizeof(slot_descs[0]);
    ObjectPool pool;
    create_slot_descriptors(&pool, &src_slot_descriptors, slot_descs, n);

    {
        OrcScannerAdapter adapter(src_slot_descriptors);
        auto input_stream = orc::readLocalFile(input_orc_file);
        Status st = adapter.init(std::move(input_stream));
        DCHECK(st.ok()) << st.get_error_msg();

        st = adapter.read_next();
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr ckptr = adapter.create_chunk();
        DCHECK(ckptr != nullptr);
        st = adapter.fill_chunk(&ckptr);
        DCHECK(st.ok()) << st.get_error_msg();
        ChunkPtr result = adapter.cast_chunk(&ckptr);
        DCHECK(result != nullptr);

        EXPECT_EQ(result->num_rows(), 2);
        EXPECT_EQ(result->num_columns(), 4);

        for (int i = 0; i < 4; i++) {
            ColumnPtr col = result->get_column_by_slot_id(i);
            std::cout << "column" << i << ": " << col->debug_string() << std::endl;
        }
    }
}

} // namespace starrocks::vectorized
